//-----------------------------------------------------------------------
// <copyright file="Texify.cs" company="Cord Burmeister">
//     Copyright  2012 Cord Burmeister. All rights reserved.
// 	Microsoft Public License (http://www.microsoft.com/opensource/licenses.mspx#Ms-PL)
// </copyright>
// <authorCord Burmeister</author>
// <email>danctex2010@outlook.com</email>
//-----------------------------------------------------------------------
namespace DancTeX.Build.Tasks
{
    using System;
    using System.Collections.Generic;
    using System.Globalization;
    using System.IO;
    using System.Text.RegularExpressions;
    using DancTeX.Build.Tasks;
    using Microsoft.Build.Framework;
    using Microsoft.Build.Utilities;

    /// <summary>
    /// An MSBuild task to run the LaTeX compiler.
    /// </summary>
    [Serializable]
//    [LoadInSeparateAppDomain]
    public sealed class Texify : ToolTask, IDisposable
    {
        /// <summary>
        /// The list of files which are stacked up during paring the LaTeX source
        /// </summary>
        private readonly Stack<string> fileNameStack = new Stack<string>();

        /// <summary>
        /// List of errors pares by the tasks
        /// </summary>
        private readonly List<DancTeXErrorInfo> errorList = new List<DancTeXErrorInfo>();

        /// <summary>
        /// This is the string which identifies the start of the LaTeX output
        /// </summary>
        private const string PdfToolStartIdentification = "This is pdfTeX, Version 3.1415926"; // ....
        
        /// <summary>
        /// This is the string which identifies the end of the LaTeX output
        /// </summary>
        private const string PdfToolEndIdentification = "Output written on"; // ....

        /// <summary>
        /// This is the string which identifies the start of the pdfTeX output
        /// </summary>
        private const string DviToolStartIdentification = "This is pdfTeX, Version 3.1415926"; // ....

        /// <summary>
        /// This is the string which identifies the end of the pdfTeX output
        /// </summary>
        private const string DviToolEndIdentification = "Output written on"; // ....

        /// <summary>
        /// This is the name of the compiler utility to perform all required actions
        /// </summary>
        private const string TexifyToolName = "texify.exe";

        /// <summary>
        /// Bool if we are parins inside the the LaTeX output-
        /// </summary>
        private bool isInLaTeXOutput;

        /// <summary>
        /// remember the output
        /// </summary>
        private string rememberedOutput = string.Empty;

        /// <summary>
        /// Remember if anything is disposed
        /// </summary>
        private bool disposed;

        /// <summary>
        /// List of intermediate files
        /// </summary>
        private List<ITaskItem> intermediateFiles = new List<ITaskItem>();

        /// <summary>
        /// This indicated the type of the output. PDF or DVI 
        /// </summary>
        private ITaskItem outputType;

        /// <summary>
        /// Gets or sets IntermediateFiles
        /// </summary>        /// <value>The output file.</value>
        [Output]
        public ITaskItem[] IntermediateFiles
        {
            get
            {
                return this.intermediateFiles.ToArray();
            }

            set
            {
                this.intermediateFiles = new List<ITaskItem>(value);
            }
        }

        /// <summary>
        /// Gets or sets OutputType
        /// </summary>        /// <value>The type of the output.</value>
        public ITaskItem OutputType
        {
            get
            {
                return this.outputType;
            }

            set
            {
                this.outputType = value;
            }
        }

        /// <summary>
        /// Gets or sets the source files.
        /// </summary>
        /// <value>The source files.</value>
        [Required]
        public ITaskItem[] SourceFiles { get; set; }

        /// <summary>
        /// Gets the name of the executable.
        /// </summary>
        /// <value>The name of the executable.</value>
        /// <remarks>The ToolName is used with the ToolPath to get the location of candle.exe.</remarks>
        protected override string ToolName
        {
            get { return TexifyToolName; }
        }

        /// <summary>
        /// Cleans up the Texify members
        /// </summary>
        /// <exception cref="System.ArgumentNullException">Thrown when SuppressFinalize throws exception. </exception>
        public void Dispose()
        {
            if (this.disposed)
            {
                return;
            }

            this.Dispose(true);
            GC.SuppressFinalize(this);
            this.disposed = true;
        }

        /// <summary>
        /// Runs the exectuable file with the specified task parameters.
        /// </summary>
        /// <returns>
        /// true if the task runs successfully; otherwise, false.
        /// </returns>
        public override bool Execute()
        {
            bool success = base.Execute();

            this.EjectErrors();

            if (success)
            {
                foreach (ITaskItem taskItem in this.SourceFiles)
                {
                    string fileName = taskItem.ToString();
                    string baseOutputFileName;
                    string outputFileName;
                    baseOutputFileName = Path.Combine(Path.GetDirectoryName(fileName), Path.GetFileName(fileName));

                    outputFileName = baseOutputFileName + ".aux";
                    if (File.Exists(outputFileName))
                    {
                        TaskItem newOutputFile = new TaskItem();
                        newOutputFile.ItemSpec = outputFileName;
                        this.intermediateFiles.Add(newOutputFile);
                    }

                    outputFileName = baseOutputFileName + ".log";
                    if (File.Exists(outputFileName))
                    {
                        TaskItem newOutputFile = new TaskItem();
                        newOutputFile.ItemSpec = outputFileName;
                        this.intermediateFiles.Add(newOutputFile);
                    }

                    outputFileName = baseOutputFileName + ".out";
                    if (File.Exists(outputFileName))
                    {
                        TaskItem newOutputFile = new TaskItem();
                        newOutputFile.ItemSpec = outputFileName;
                        this.intermediateFiles.Add(newOutputFile);
                    }

                    outputFileName = baseOutputFileName + ".toc";
                    if (File.Exists(outputFileName))
                    {
                        TaskItem newOutputFile = new TaskItem();
                        newOutputFile.ItemSpec = outputFileName;
                        this.intermediateFiles.Add(newOutputFile);
                    }
                }
            }

            return success;
        }

        /// <summary>
        /// Get the path to the executable. 
        /// </summary>
        /// <remarks>GetFullPathToTool is only called when the ToolPath property is not set (see the ToolName remarks above).</remarks>
        /// <returns>The full path to the executable or simply candle.exe if it's expected to be in the system path.</returns>
        /// <exception cref="System.ArgumentException">Thrown when argument exception occurs. </exception>
        /// <exception cref="System.ArgumentNullException">Thrown when null arguments are passed. </exception>
        /// <exception cref="System.IO.PathTooLongException">Thrown when path name is too long </exception>
        /// <exception cref="System.NotSupportedException">Thrown when not supported argeumtns are passed.</exception>
        /// <exception cref="System.Security.SecurityException">Thrown when security exception occurs </exception>
        protected override string GenerateFullPathToTool()
        {
            try
            {
                // Removed that section , because it makes the call unstable. 
                // Proting ToolTask to own Appdomain might help.
                //MiKTeXSession2_9.ISession2 session = new MiKTeXSession2_9.MiKTeXSession2_9Class();
                //MiKTeXSession2_9.MiKTeXSetupInfo setupInfo = session.GetMiKTeXSetupInfo();
                //ToolPath = setupInfo.binDirectory;
                //ToolPath = @"C:\Program Files (x86)\MiKTeX 2.9\miktex\bin";            
            }
            catch
            {
                // nothing to do 
                throw;
            }

            // If there's not a ToolPath specified, it has to be in the system path.
            if (String.IsNullOrEmpty(ToolPath))
            {
                return TexifyToolName;
            }

            return Path.Combine(Path.GetFullPath(ToolPath), TexifyToolName);
        }

        /// <summary>
        /// Returns a string value containing the command line arguments to pass directly to the executable file.
        /// </summary>
        /// <returns>
        /// A string value containing the command line arguments to pass directly to the executable file.
        /// </returns>
        protected override string GenerateCommandLineCommands()
        {
            DancTeXCommandLineBuilder commandLineBuilder = new DancTeXCommandLineBuilder();
            this.BuildCommandLine(commandLineBuilder);
            return commandLineBuilder.ToString();
        }

        /// <summary>
        /// Logs the tool name and path to all registered loggers.
        /// </summary>
        /// <param name="toolName">The executable file name.</param>
        /// <param name="pathToTool">The path to the executable file.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when null argeuments are passed. </exception>
        protected override void LogPathToTool(string toolName, string pathToTool)
        {
            this.Log.LogMessage(MessageImportance.High, pathToTool);
            base.LogPathToTool(toolName, pathToTool);
        }

        /// <summary>
        /// Parses a single line of text to identify any errors or warnings in canonical format.
        /// </summary>
        /// <param name="singleLine">A single line of text for the method to parse.</param>
        /// <param name="messageImportance">A value of <see cref="T:Microsoft.Build.Framework.MessageImportance"/> that indicates the importance level with which to log the message.</param>
        /// <exception cref="System.ArgumentNullException">Thrown when arguments are null. </exception>
        protected override void LogEventsFromTextOutput(string singleLine, MessageImportance messageImportance)
        {
            this.ParseToolOutputLine(singleLine);
            this.Log.LogMessage(MessageImportance.High, singleLine);

            // we handle the error forwarding completly outselfs
            // base.LogEventsFromTextOutput(singleLine, MessageImportance.High);
        }

        /// <summary>
        /// Cleans up the ManualResetEvent members
        /// </summary>
        /// <param name="disposing">if set to <c>true</c> [disposing].</param>
        private void Dispose(bool disposing)
        {
            if (disposing)
            {
                this.fileNameStack.Clear();
            }
        }

        /// <summary>
        /// Builds a command line from options in this task.
        /// </summary>
        /// <param name="commandLineBuilder">The command line builder.</param>
        private void BuildCommandLine(CommandLineBuilder commandLineBuilder)
        {
            if ((this.outputType != null) && (this.outputType.ToString() == "PDF"))
            {
                commandLineBuilder.AppendSwitch("-p");
            }

            // Add the sourceing information to the DVI File if possible 
            commandLineBuilder.AppendSwitch("--tex-option=--src");

            commandLineBuilder.AppendFileNamesIfNotNull(this.SourceFiles, " ");
        }

        /// <summary>
        /// Parses the tool output line.
        /// </summary>
        /// <param name="line">The line of output from the LaTeX program.</param>
        private void ParseToolOutputLine(string line)
        {
            if ((this.outputType != null) && (this.outputType.ToString() == "PDF"))
            {
                if (!this.isInLaTeXOutput)
                {
                    if (line.StartsWith(PdfToolStartIdentification, StringComparison.Ordinal))
                    {
                        this.ResetErrorCollection();

                        // this.Log.LogError();                       
                        this.isInLaTeXOutput = true;
                    }
                }

                if (this.isInLaTeXOutput)
                {
                    if (line.StartsWith(PdfToolEndIdentification, StringComparison.Ordinal))
                    {
                        this.isInLaTeXOutput = false;
                    }
                }
            }

            if ((this.outputType != null) && (this.outputType.ToString() == "DVI"))
            {
                if (!this.isInLaTeXOutput)
                {
                    if (line.StartsWith(DviToolStartIdentification, StringComparison.Ordinal))
                    {
                        this.ResetErrorCollection();

                        // this.Log.LogError();                       
                        this.isInLaTeXOutput = true;
                    }
                }

                if (this.isInLaTeXOutput)
                {
                    if (line.StartsWith(DviToolEndIdentification, StringComparison.Ordinal))
                    {
                        this.isInLaTeXOutput = false;
                    }
                }
            }

            if (this.isInLaTeXOutput)
            {
                // Scan for file names in the output 
                string trimmedLine = line.Trim();
                if (!string.IsNullOrEmpty(this.rememberedOutput))
                {
                    trimmedLine = this.rememberedOutput + trimmedLine;
                    this.rememberedOutput = string.Empty;
                }

                Match m;                

                #region Warning: Reference
                // LaTeX Warning: Reference `brief' on page 42 undefined on input line 165.
                m = Regex.Match(trimmedLine, @"LaTeX Warning: Reference \`[a-zA-Z0-9]+\' on page [0-9]+ undefined on input line (?<lineNumber>[0-9]+)\.");
                if (m.Success)
                {
                    string fileName = this.fileNameStack.Peek();
                    int lineNumber = Int32.Parse((string)m.Groups["lineNumber"].ToString(), CultureInfo.InvariantCulture);
                    this.errorList.Add(
                        new DancTeXErrorInfo(DancTeXErrorInfoType.Warning, lineNumber, lineNumber, trimmedLine, fileName));
                    return;
                }
                #endregion

                #region Warning:  Citation
                m = Regex.Match(trimmedLine, @"LaTeX Warning: Citation \`[a-zA-Z0-9]+\' on page [0-9]+ undefined on input line (?<lineNumber>[0-9]+)\.");
                if (m.Success == true)
                {
                    string fileName = this.fileNameStack.Peek();
                    int lineNumber = Int32.Parse((string)m.Groups["lineNumber"].ToString(), CultureInfo.InvariantCulture);
                    this.errorList.Add(
                        new DancTeXErrorInfo(DancTeXErrorInfoType.Warning, lineNumber, lineNumber, trimmedLine, fileName));
                    return;
                }
                #endregion

                #region Handle overfull hbox warning
                // Handle overfull hbox warning
                m = Regex.Match(trimmedLine, @"Overfull \\hbox \(.*\) in paragraph at lines (?<FromLine>[0-9]+)\-\-(?<ToLine>[0-9]+)");
                if (m.Success == true)
                {
                    string fileName = this.fileNameStack.Peek();
                    int fromLine = Int32.Parse((string)m.Groups["FromLine"].ToString(), CultureInfo.InvariantCulture);
                    int toLine = Int32.Parse((string)m.Groups["ToLine"].ToString(), CultureInfo.InvariantCulture);
                    this.errorList.Add(
                        new DancTeXErrorInfo(DancTeXErrorInfoType.Warning, fromLine, toLine, trimmedLine, fileName));
                    return;
                } 
                #endregion

                #region Handle Underfull  hbox warning
                // Handle overfull hbox warning
                m = Regex.Match(trimmedLine, @"Underfull \\hbox \(.*\) in paragraph at lines (?<FromLine>[0-9]+)\-\-(?<ToLine>[0-9]+)");
                if (m.Success == true)
                {
                    string fileName = this.fileNameStack.Peek();
                    int fromLine = Int32.Parse((string)m.Groups["FromLine"].ToString(), CultureInfo.InvariantCulture);
                    int toLine = Int32.Parse((string)m.Groups["ToLine"].ToString(), CultureInfo.InvariantCulture);
                    this.errorList.Add(
                        new DancTeXErrorInfo(DancTeXErrorInfoType.Warning, fromLine, toLine, trimmedLine, fileName));
                    return;
                }
                #endregion

                #region Handle  Undefined control sequence error
                // Handle  Undefined control sequence error
                m = Regex.Match(trimmedLine, @"\! Undefined control sequence\..*l\.(?<lineNumber>[0-9]+)\s+(?<controlSequence>\\[a-zA-Z0-9]+)");
                if (m.Success == true)
                {
                    string fileName = this.fileNameStack.Peek();
                    Console.WriteLine("Error and fileName " + fileName);
                    int lineNumber = Int32.Parse((string)m.Groups["lineNumber"].ToString(), CultureInfo.InvariantCulture);

                    this.errorList.Add(
                        new DancTeXErrorInfo(DancTeXErrorInfoType.Error, lineNumber, lineNumber, "Undefined control sequence " + m.Groups["controlSequence"].ToString(), fileName));
                    return;
                }

                // Handle  Undefined control sequence error uncomplete error message
                m = Regex.Match(trimmedLine, @"\! Undefined control sequence.");
                if (m.Success == true)
                {
                    this.rememberedOutput = trimmedLine;
                    return;
                }
                #endregion

                #region Handle too many '}' 
                // Handle  Handle too many '}'
                m = Regex.Match(trimmedLine, @"\! Too many \}\'s\..*l\.(?<lineNumber>[0-9]+)");
                if (m.Success == true)
                {
                    string fileName = this.fileNameStack.Peek();
                    int lineNumber = Int32.Parse((string)m.Groups["lineNumber"].ToString(), CultureInfo.InvariantCulture);

                    this.errorList.Add(
                        new DancTeXErrorInfo(DancTeXErrorInfoType.Error, lineNumber, lineNumber, "Too many }'s ", fileName));
                    return;
                }

                // Handle  Undefined control sequence error uncomplete error message
                m = Regex.Match(trimmedLine, @"\! Too many \}\'s\.");
                if (m.Success == true)
                {
                    this.rememberedOutput = trimmedLine;
                    return;
                }
                #endregion

                #region ! LaTeX Warning:

                // Handle  Handle too many '}'
                m = Regex.Match(trimmedLine, @"\! LaTeX Warning: (?<errorMessage>[^\.]+).*l\.(?<lineNumber>[0-9]+)");
                if (m.Success == true)
                {
                    string fileName = this.fileNameStack.Peek();
                    int lineNumber = Int32.Parse((string)m.Groups["lineNumber"].ToString(), CultureInfo.InvariantCulture);

                    this.errorList.Add(
                        new DancTeXErrorInfo(DancTeXErrorInfoType.Warning, lineNumber, lineNumber, m.Groups["errorMessage"].ToString() + ".", fileName));
                    return;
                }

                // Handle  Undefined control sequence error uncomplete error message
                m = Regex.Match(trimmedLine, @"\! LaTeX Warning:");
                if (m.Success == true)
                {
                    this.rememberedOutput = trimmedLine;
                    return;
                }

                #endregion

                #region ! LaTeX Error:

                // Handle  Handle too many '}'
                m = Regex.Match(trimmedLine, @"\! LaTeX Error: (?<errorMessage>[^\.]+).*l\.(?<lineNumber>[0-9]+)");
                if (m.Success == true)
                {
                    string fileName = this.fileNameStack.Peek();
                    int lineNumber = Int32.Parse((string)m.Groups["lineNumber"].ToString(), CultureInfo.InvariantCulture);

                    this.errorList.Add(
                        new DancTeXErrorInfo(DancTeXErrorInfoType.Error, lineNumber, lineNumber, m.Groups["errorMessage"].ToString() + ".", fileName));
                    return;
                }

                // Handle  Undefined control sequence error uncomplete error message
                m = Regex.Match(trimmedLine, @"\! LaTeX Error:");
                if (m.Success == true)
                {
                    this.rememberedOutput = trimmedLine;
                    return;
                }

                #endregion

                #region Handle Pdftex warnings
                // pdfTeX warning (ext4):
                m = Regex.Match(trimmedLine, @"(?<pdfMessagePrefix>.*)(?<pdfWarning>pdfTeX warning.*: .*)l\.(?<lineNumber>[0-9]+)\s+(?<controlSequence>.+)");
                if (m.Success == true)
                {
                    string prefix = m.Groups["pdfMessagePrefix"].ToString();
                    if (!string.IsNullOrEmpty(prefix))
                    {
                        this.ParsePageOutputLine(prefix);
                    }

                    string fileName = this.fileNameStack.Peek();
                    Console.WriteLine("Error and fileName " + fileName);
                    int lineNumber = Int32.Parse((string)m.Groups["lineNumber"].ToString(), CultureInfo.InvariantCulture);

                    this.errorList.Add(
                        new DancTeXErrorInfo(DancTeXErrorInfoType.Warning, lineNumber, lineNumber, m.Groups["pdfWarning"].ToString(), fileName));
                    return;
                }

                // Handle  pdfTeX warning uncomplete error message
                m = Regex.Match(trimmedLine, @".*pdfTeX warning.*:");
                if (m.Success == true)
                {
                    this.rememberedOutput = trimmedLine;
                    return;
                }

                #endregion 

                #region Handle Pdftex errors
                // pdfTeX warning (ext4):
                m = Regex.Match(trimmedLine, @"(?<pdfMessagePrefix>.*)(?<pdfWarning>pdfTeX error.*: .*)l\.(?<lineNumber>[0-9]+)\s+(?<controlSequence>.+)");
                if (m.Success == true)
                {
                    string prefix = m.Groups["pdfMessagePrefix"].ToString();
                    if (!string.IsNullOrEmpty(prefix))
                    {
                        this.ParsePageOutputLine(prefix);
                    }

                    string fileName = this.fileNameStack.Peek();
                    Console.WriteLine("Error and fileName " + fileName);

                    int lineNumber = Int32.Parse((string)m.Groups["lineNumber"].ToString(), CultureInfo.InvariantCulture);

                    this.errorList.Add(
                        new DancTeXErrorInfo(DancTeXErrorInfoType.Error, lineNumber, lineNumber, m.Groups["pdfWarning"].ToString(), fileName));
                    return;
                }

                // Handle  pdfTeX warning uncomplete error message
                m = Regex.Match(trimmedLine, @".*pdfTeX error.*:");
                if (m.Success == true)
                {
                    this.rememberedOutput = trimmedLine;
                    return;
                }

                #endregion 

                #region Handle  ! Emergency stop.
                // ! Emergency stop.
                m = Regex.Match(trimmedLine, @"\! Emergency stop\..*l\.(?<lineNumber>[0-9]+)");
                if (m.Success == true)
                {
                    string fileName = this.fileNameStack.Peek();
                    int lineNumber = Int32.Parse((string)m.Groups["lineNumber"].ToString(), CultureInfo.InvariantCulture);

                    this.errorList.Add(
                        new DancTeXErrorInfo(DancTeXErrorInfoType.Error, lineNumber, lineNumber, "Emergency stop error.", fileName));
                    return;
                }

                // Handle  Undefined control sequence error uncomplete error message
                m = Regex.Match(trimmedLine, @"\! Emergency stop\.");
                if (m.Success == true)
                {
                    this.rememberedOutput = trimmedLine;
                    return;
                }
                #endregion

                if (trimmedLine.StartsWith("(", StringComparison.Ordinal) ||
                    trimmedLine.StartsWith(")", StringComparison.Ordinal) ||
                    trimmedLine.StartsWith("]", StringComparison.Ordinal) ||
                    trimmedLine.StartsWith("[", StringComparison.Ordinal))
                {
                    this.ParsePageOutputLine(trimmedLine);
                }
            }
        }

        /// <summary>
        /// Parses the page output line.
        /// </summary>
        /// <param name="line">The line to be parsed.</param>
        private void ParsePageOutputLine(string line)
        {
           string trimmedLine = line.Trim();
           if (trimmedLine.StartsWith("(", StringComparison.Ordinal))
           {
               string saveString = trimmedLine;
               trimmedLine = trimmedLine.Remove(0, 1); // Remove first one
               if (trimmedLine.StartsWith("\"", StringComparison.Ordinal))
               {                  
                   // This is a quoted line 
                   trimmedLine = trimmedLine.Remove(0, 1); // remove first quote
                   int index = trimmedLine.IndexOf("\"", StringComparison.Ordinal); // find next quote
                   if (index > 0)
                   {
                       string fileName = trimmedLine.Substring(0, index);
                       trimmedLine = trimmedLine.Remove(0, index);
                       this.fileNameStack.Push(fileName);
                   }
                   else
                   {
                       this.rememberedOutput = saveString;
                       return;
                   }
               }
               else
               {
                   char[] delimiter = { ' ', ')', '(', '[', ']' };
                   string[] splitted = trimmedLine.Split(delimiter);
                   if (splitted.Length > 0)
                   {
                       string fileName = splitted[0];
                       trimmedLine = trimmedLine.Remove(0, fileName.Length);
                       this.fileNameStack.Push(fileName);
                   }
                   else
                   {
                       this.rememberedOutput = saveString;
                       return;
                   }
               }
           }
           else if (trimmedLine.StartsWith(")", StringComparison.Ordinal))
           {
               trimmedLine = trimmedLine.Remove(0, 1);
               if (this.fileNameStack.Count > 0)
               {
                   this.fileNameStack.Pop();
               }

               if (this.fileNameStack.Count > 0)
               {
                   string fileName = this.fileNameStack.Peek();
               }
               else
               {
                   // this.Log.LogWarning("DONE", "fileName", "helpkeyword", "fileName", 1, 1, 1, 1, trimmedLine, null);
               }
           }
           else
           {
               trimmedLine = trimmedLine.Remove(0, 1);
           }

           if (trimmedLine.Length > 0)
           {
               this.ParsePageOutputLine(trimmedLine);
           }
        }

        /// <summary>
        /// Resets the error collection.
        /// </summary>
        private void ResetErrorCollection()
        {
            this.errorList.Clear();
        }

        /// <summary>
        /// Ejects the errors.
        /// </summary>
        private void EjectErrors()
        {
            foreach (DancTeXErrorInfo errorInfo in this.errorList)
            {
                if (errorInfo.Type == DancTeXErrorInfoType.Error)
                {
                    this.Log.LogError(
                        "subcategory", "1000", "helpkeyword", errorInfo.FileName, errorInfo.FromLine, 0, errorInfo.ToLine, 0, errorInfo.Message, null);
                }
                else
                {
                    this.Log.LogWarning(
                        "subcategory", "1000", "helpkeyword", errorInfo.FileName, errorInfo.FromLine, 0, errorInfo.ToLine, 0, errorInfo.Message, null);
                }
            }
        }
    }
}